""" A Generalized Linear Integrate and Fire (GLIF) neuron modeling package. 
Use this code to run the GLIF models available in the Allen Cell Types Atlas.
See :doc:`glif_models` for more details.
"""
